"""Coiled cluster information displayed in a Jupyter Notebook-like context."""
from __future__ import annotations

import uuid
from collections import defaultdict
from datetime import datetime
from pathlib import Path
from types import TracebackType
from typing import Iterable, Type

import jmespath
from IPython.display import DisplayHandle, clear_output, display
from jinja2 import Environment, FileSystemLoader, Template
from jinja2.exceptions import TemplateNotFound

from ...errors import ClusterCreationError
from ..states import (
    DISPLAY_DATETIME_FORMAT,
    CombinedProcessStateEnum,
    ProcessStateEnum,
    State,
    StatefulObjectType,
)
from .util import get_instance_types, get_worker_statuses

try:
    import ipywidgets as ipyw
except ImportError:
    pass


def format_state(state: State) -> str:
    """Format the state of a worker or scheduler consistently.

    The output is intended for the status logs views in the cluster status widgets.
    """
    state_names = {
        StatefulObjectType.cluster: "Cluster",
        StatefulObjectType.scheduler: "Scheduler Process",
        StatefulObjectType.scheduler_instance: "Scheduler Instance",
        StatefulObjectType.cluster_infra: "Network Infrastructure",
        StatefulObjectType.worker: "Worker Process",
        StatefulObjectType.worker_instance: "Worker Instance",
    }
    if state.type in (StatefulObjectType.worker, StatefulObjectType.worker_instance):
        max_type_length = 15
    elif state.type in (
        StatefulObjectType.scheduler,
        StatefulObjectType.scheduler_instance,
    ):
        max_type_length = 18
    else:
        max_type_length = max(len(v) for v in state_names.values())
    time = datetime.strftime(state.updated.astimezone(), DISPLAY_DATETIME_FORMAT)
    return (
        f"{state_names[state.type]:<{max_type_length}} | {state.state:<8} | {time} "
        f"| {state.reason}"
    )


def format_error_state(state: State) -> str:
    """Format the error state of a worker or scheduler consistently.

    The output is intended for the status logs views in the cluster status widgets.
    """
    state_names = {
        StatefulObjectType.cluster: "Cluster",
        StatefulObjectType.scheduler: "Scheduler Process",
        StatefulObjectType.scheduler_instance: "Scheduler Instance",
        StatefulObjectType.cluster_infra: "Network Infrastructure",
        StatefulObjectType.worker: "Worker Process",
        StatefulObjectType.worker_instance: "Worker Instance",
    }
    max_type_length = max(len(v) for v in state_names.values())
    time = datetime.strftime(state.updated.astimezone(), DISPLAY_DATETIME_FORMAT)
    return f"{state_names[state.type]:<{max_type_length}} | {time} | {state.reason}"


TEMPLATE_PATHS = [Path(__file__).parent / "templates"]
FILTERS = {
    "format_state": format_state,
    "format_error_state": format_error_state,
}


def get_environment() -> Environment:
    """Create a Jinja Environment instance."""
    loader = FileSystemLoader(TEMPLATE_PATHS)
    environment = Environment(loader=loader)
    environment.filters.update(FILTERS)

    return environment


def get_template(name: str) -> Template:
    """Load a Jinja template."""
    try:
        return get_environment().get_template(name)
    except TemplateNotFound as e:
        raise TemplateNotFound(
            f"Unable to find {name} in coiled.widgets.TEMPLATE_PATHS {TEMPLATE_PATHS}"
        ) from e


def _update_cluster_status(cluster_status: Iterable[State]) -> dict[str, list[State]]:
    """Separates cluster status into error, scheduler, and worker buckets.

    Each bucket corresponds to a key in the return dictionary. Errors are collected for
    all the object types, whereas the scheduler and worker keys group status messages
    for the * and *_instance types.
    """
    output = defaultdict(list)
    for state in cluster_status:
        if state.state == "error":
            output["error"].append(state)
        elif state.type in (
            StatefulObjectType.scheduler,
            StatefulObjectType.scheduler_instance,
        ):
            output["scheduler"].append(state)
        elif state.type in (
            StatefulObjectType.worker,
            StatefulObjectType.worker_instance,
        ):
            output["worker"].append(state)
    return output


CLOUDS: dict[str | None, str] = {
    "vm_aws": "Amazon Web Services",
    "vm_gcp": "Google Cloud Platform",
    None: "Unknown",
}

PUBLIC_IP_ADDRESS = jmespath.compile("scheduler.instance.public_ip_address")


class ClusterStatusWidget:  # pragma: no cover
    """Shows the status of a Coiled cluster using ipywidgets.

    The format is handled by Jinja templates in the local 'templates' directory.

    This widget is not recommended for use at this time. It's functional but not
    polished.
    """

    _display_id: DisplayHandle | None

    def __init__(self):
        """Set up the top-level widget and the display machinery."""
        self._display_id = None
        self._received_data = False
        self.tab_box = ipyw.Tab()

        # Compatibility for ipywidgets 7/8
        if hasattr(self.tab_box, "titles"):
            self.tab_box.titles = ("Overview", "Configuration", "Status Logs")
        else:
            self.tab_box.set_title(0, "Overview")
            self.tab_box.set_title(1, "Configuration")
            self.tab_box.set_title(2, "Status Logs")

        self.scheduler_status_logs = []
        self.worker_status_logs = []
        self.error_logs = []

        self.overview_tab_template = get_template("overview.html.j2")
        self.overview_tab = ipyw.HTML()

        self.configuration_template = get_template("configuration.html.j2")
        self.configuration_tab = ipyw.HTML()

        self.status_logs_template = get_template("status_logs.html.j2")
        self.status_output_tab = ipyw.HTML()

        self.tab_box.children = (
            self.overview_tab,
            self.configuration_tab,
            self.status_output_tab,
        )

    def __enter__(self) -> ClusterStatusWidget:
        """Enter a live-updating context.

        Example
        -------
        with ClusterStatusWidget() as w:
            # do stuff
            w.update(cluster_details, cluster_status)
        """
        self._display_id = display_id = DisplayHandle()
        display_id.display(self)
        return self

    def __exit__(
        self,
        exc_type: Type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exit the live-updating context and reset the display ID.

        Keep the widget around for user inspection if there was a cluster creation
        error, otherwise remove it.
        """
        if exc_type != ClusterCreationError:
            clear_output(wait=True)
        self._display_id = None

    def _ipython_display_(self):
        if self._display_id:
            # Don't bother displaying if we haven't set the data yet
            if self._received_data:
                self._display_id.display(self.tab_box)
        else:
            display(self.tab_box)

    def update(self, cluster_details, cluster_status):
        """Update the widget display data."""
        # Overview tab
        n_workers = len(jmespath.search("workers[*]", cluster_details) or [])
        statuses = get_worker_statuses(cluster_details)
        overall_cluster_status = jmespath.search("current_state.state", cluster_details)
        current_scheduler_state = jmespath.search(
            "scheduler.current_state.state", cluster_details
        )
        scheduler_ready = (
            ProcessStateEnum(current_scheduler_state) == ProcessStateEnum.started
        )
        dashboard_address = (
            jmespath.search("scheduler.dashboard_address", cluster_details)
            if scheduler_ready
            else None
        )

        status_logs = _update_cluster_status(cluster_status)
        scheduler_status_logs = status_logs["scheduler"]
        worker_status_logs = status_logs["worker"]
        error_logs = status_logs["error"]

        n_starting = (
            statuses[CombinedProcessStateEnum.instance_queued]
            + statuses[CombinedProcessStateEnum.instance_starting]
        )
        self.overview_tab.value = self.overview_tab_template.render(
            cluster_name=cluster_details["name"],
            overall_cluster_status=overall_cluster_status,
            dashboard_address=dashboard_address,
            n_workers=n_workers,
            n_starting=n_starting,
            n_running=statuses[CombinedProcessStateEnum.instance_running],
            n_ready=statuses[CombinedProcessStateEnum.ready],
            n_errored=statuses[CombinedProcessStateEnum.error],
            error_logs=error_logs,
        )

        # Configuration tab
        scheduler_instance_type, worker_instance_types = get_instance_types(
            cluster_details
        )
        # TODO: This will awkwardly wrap to the next line in the output if there are
        # too many worker types. It'd be nice to format that more nicely.
        worker_instance_types_label = ", ".join(
            f"{k} ({v:,})" for k, v in worker_instance_types.items()
        )
        worker_instance_types_label = worker_instance_types_label or None
        # TODO: This value is not yet deployed to the staging API
        software_environment_name = None
        cloud_name = CLOUDS.get(
            jmespath.search("cluster_options.type", cluster_details), "Unknown"
        )
        region_name = jmespath.search("cluster_options.region_name", cluster_details)
        self.configuration_tab.value = self.configuration_template.render(
            scheduler_instance_type=scheduler_instance_type,
            worker_instance_type=worker_instance_types_label,
            software_environment_name=software_environment_name,
            cloud_name=cloud_name,
            region_name=region_name,
        )

        # Status logs tab
        self.status_output_tab.value = self.status_logs_template.render(
            scheduler_status_logs=scheduler_status_logs,
            worker_status_logs=worker_status_logs,
        )

        # Handle an initial display
        if not self._received_data:
            self._received_data = True
            if self._display_id:
                self._display_id.display(self.tab_box)


class HTMLClusterWidget:
    """Shows the status of a Coiled cluster using plain HTML.

    The format is handled by Jinja templates in the local 'templates' directory.
    """

    _display_id: DisplayHandle | None

    def __init__(self) -> None:
        """Set up the template and other machinery."""
        self._display_id = None
        self._received_data = False
        self.template = get_template("cluster_status.html.j2")
        self.scheduler_status_logs = []
        self.worker_status_logs = []
        self.error_logs = []

    def __enter__(self) -> HTMLClusterWidget:
        """Enter a live-updating context.

        Example
        -------
        with HTMLClusterWidget() as w:
            # do stuff
            w.update(cluster_details, cluster_status)
        """
        self._display_id = display_id = DisplayHandle()
        display_id.display(self)
        return self

    def __exit__(
        self,
        exc_type: Type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exit the live-updating context and reset the display ID.

        Update the widget to the interactive state. Keep the widget around for user
        inspection if there was a cluster creation error, otherwise remove it.
        """
        if self._display_id and self._received_data:
            mimebundle = {
                "text/html": self.template.render(
                    interactive=True, unique=uuid.uuid4(), **self.__dict__
                )
            }
            self._display_id.update(mimebundle, raw=True)

        # If we got a cluster creation error, let the output hang around
        # for user inspection.
        if exc_type != ClusterCreationError:
            clear_output(wait=True)
        self._display_id = None

    def _ipython_display_(self):
        # Don't bother displaying if we haven't set the data.
        if self._received_data:
            interactive = not self._display_id
            mimebundle = {
                "text/html": self.template.render(
                    interactive=interactive, unique=uuid.uuid4(), **self.__dict__
                )
            }
            if self._display_id:
                self._display_id.display(mimebundle, raw=True)
            else:
                display(mimebundle, raw=True)

    def update(self, cluster_details, cluster_status, *args, **kwargs):
        """Update the widget display data."""
        self.cluster_name = jmespath.search("name", cluster_details)
        self.cloud_name = CLOUDS.get(
            jmespath.search("cluster_options.type", cluster_details), "Unknown"
        )
        self.region_name = jmespath.search(
            "cluster_options.region_name", cluster_details
        )
        self.software_environment_name = None
        self.overall_cluster_status = jmespath.search(
            "current_state.state", cluster_details
        )
        self.scheduler_status = jmespath.search(
            "scheduler.current_state.state", cluster_details
        )
        scheduler_ready = (
            ProcessStateEnum(self.scheduler_status) == ProcessStateEnum.started
        )
        self.dashboard_address = (
            jmespath.search("scheduler.dashboard_address", cluster_details)
            if scheduler_ready
            else None
        )
        self.n_workers = len(jmespath.search("workers[*]", cluster_details) or [])

        statuses = get_worker_statuses(cluster_details)
        self.n_starting = (
            statuses[CombinedProcessStateEnum.instance_queued]
            + statuses[CombinedProcessStateEnum.instance_starting]
        )
        self.n_running = statuses[CombinedProcessStateEnum.instance_running]
        self.n_ready = statuses[CombinedProcessStateEnum.ready]
        self.n_errored = statuses[CombinedProcessStateEnum.error]
        status = _update_cluster_status(cluster_status)
        self.scheduler_status_logs = status["scheduler"]
        self.worker_status_logs = status["worker"]
        self.error_logs = status["error"]
        # TODO: Add software_environment_name into this check/update as well
        self.scheduler_instance_type, worker_instance_types = get_instance_types(
            cluster_details
        )
        self.worker_instance_type = ", ".join(
            f"{k} ({v:,})" for k, v in worker_instance_types.items()
        )
        self.worker_instance_type = self.worker_instance_type or None

        # Include a UUID so that the bundled script, which refers to elements by ID,
        # grabs only the one that is rendered here, rather than any others that might
        # be on the page.
        interactive = not self._display_id
        mimebundle = {
            "text/html": self.template.render(
                interactive=interactive, unique=uuid.uuid4(), **self.__dict__
            )
        }

        if self._display_id:
            # Handle an initial display
            if not self._received_data:
                self._display_id.display(mimebundle, raw=True)
            else:
                self._display_id.update(mimebundle, raw=True)
        self._received_data = True



import asyncio
import json
import os
import platform
import sys
import typing
from logging import getLogger
from pathlib import Path
from typing import Dict, List, Optional, Union

from importlib_metadata import Distribution

from coiled.types import CondaPackage, CondaPlaceHolder, PackageInfo

logger = getLogger("coiled.package_sync")
subdir_datas = {}
PYTHON_VERSION = platform.python_version_tuple()


async def scan_conda(prefix: Path) -> typing.Dict[str, PackageInfo]:
    conda_meta = prefix / "conda-meta"
    if conda_meta.exists() and conda_meta.is_dir():
        conda_packages = [
            CondaPackage(json.load(metafile.open("r")), prefix=prefix)
            for metafile in conda_meta.iterdir()
            if metafile.suffix == ".json"
        ]
        packages = await asyncio.gather(
            *[handle_conda_package(pkg) for pkg in conda_packages]
        )
        return {pkg["name"]: pkg for pkg in packages}
    else:
        return {}


async def handle_conda_package(pkg: CondaPackage) -> PackageInfo:
    # Are there conda packages that install multiple python packages?
    metadata_location = next(
        (Path(fp).parent for fp in pkg.files if fp.endswith("METADATA")), None
    )
    if metadata_location:
        dist = Distribution.at(pkg.prefix / metadata_location)
        name = dist.metadata["Name"] or pkg.name
    else:
        name = pkg.name
    return {
        "channel": pkg.channel,
        "path": None,
        "channel_url": pkg.channel_url,
        "source": "conda",
        "conda_name": pkg.name,
        "subdir": pkg.subdir,
        "name": name,
        "version": pkg.version,
        "wheel_target": None,
    }


async def handle_dist(
    dist: Distribution, locations: List[Path]
) -> Optional[Union[PackageInfo, CondaPlaceHolder]]:
    installer = dist.read_text("INSTALLER") or ""
    installer = installer.rstrip()
    dist_path = dist._path  # type: ignore
    dist_path: Path
    if installer == "conda":
        return CondaPlaceHolder(name=dist.name)
    elif dist_path.parent.suffix == ".egg":
        return {
            "name": dist.name,
            "path": dist_path.parent,
            "source": "pip",
            "channel": None,
            "subdir": None,
            "channel_url": None,
            "conda_name": None,
            "version": dist.version,
            "wheel_target": str(dist_path.parent).rstrip(os.sep),
        }
    else:
        direct_url_metadata = dist.read_text("direct_url.json")
        if direct_url_metadata:
            url_metadata = json.loads(direct_url_metadata)
            if url_metadata.get("vcs_info"):
                vcs_info = url_metadata.get("vcs_info")
                vcs = vcs_info["vcs"]
                commit = vcs_info["commit_id"]
                url = url_metadata["url"]
                if vcs == "git":
                    # TODO: Download source + build sdist?
                    # this would allow private repos to work well
                    pip_url = f"git+{url}@{commit}"
                    return {
                        "name": dist.name,
                        "path": dist_path,
                        "source": "pip",
                        "channel": None,
                        "channel_url": None,
                        "subdir": None,
                        "conda_name": None,
                        "version": dist.version,
                        "wheel_target": pip_url,
                    }
            elif url_metadata.get("url"):
                return {
                    "name": dist.name,
                    "path": Path(url_metadata["url"].replace("file://", "")),
                    "source": "pip",
                    "channel": None,
                    "channel_url": None,
                    "subdir": None,
                    "conda_name": None,
                    "version": dist.version,
                    "wheel_target": url_metadata["url"],
                }
        egg_links = []
        for location in locations:
            if os.name == "nt":
                egg_link_pth = location / Path(dist.name).with_suffix(".egg-link")
                if egg_link_pth.is_file():
                    egg_links.append(
                        location / Path(dist.name).with_suffix(".egg-link")
                    )
            else:
                egg_link_pth = location / Path(dist.name).with_suffix(".egg-link")
                if egg_link_pth.is_file():
                    egg_links.append(
                        location / Path(dist.name).with_suffix(".egg-link")
                    )
        if egg_links:
            return {
                "name": dist.name,
                "path": dist_path.parent,
                "source": "pip",
                "channel": None,
                "channel_url": None,
                "subdir": None,
                "conda_name": None,
                "version": dist.version,
                "wheel_target": str(dist_path.parent),
            }
        return {
            "name": dist.name,
            "path": dist_path,
            "source": "pip",
            "channel": None,
            "channel_url": None,
            "subdir": None,
            "conda_name": None,
            "version": dist.version,
            "wheel_target": None,
        }


async def scan_pip(
    locations: List[Path],
) -> typing.Dict[str, Union[PackageInfo, CondaPlaceHolder]]:
    # distributions returns ALL distributions
    # even ones that are not active
    # this is a trick so we only get the distribution
    # that is last in stack
    locations = [
        location for location in locations if location.exists() and location.is_dir()
    ]
    paths: List[str] = [str(location) for location in locations]
    for location in locations:
        for fp in location.iterdir():
            if fp.suffix in [".pth", ".egg-link"]:
                for line in fp.read_text().split("\n"):
                    if line.startswith("#"):
                        continue
                    elif line.startswith(("import", "import\t")):
                        continue
                    elif line.rstrip() == ".":
                        continue
                    else:
                        p = location / Path(line.rstrip())
                        full_path = str(p.resolve())
                        if p.exists() and full_path not in paths:
                            paths.append(full_path)
    active_dists: Dict[str, Distribution] = {
        dist.name: dist for dist in Distribution.discover(path=list(paths))
    }
    dists = active_dists.values()
    return {
        pkg["name"]: pkg
        for pkg in await asyncio.gather(
            *(handle_dist(dist, locations) for dist in dists)
        )
        if pkg
    }


async def scan_prefix(
    prefix: Optional[Path] = None, locations: Optional[List[Path]] = None
) -> typing.List[PackageInfo]:
    # TODO: private conda channels
    # TODO: detect pre-releases and only set --pre flag for those packages (for conda)
    if not prefix:
        prefix = Path(sys.prefix)
    conda_env_future = asyncio.create_task(scan_conda(prefix=prefix))
    # only pass locations to support testing, otherwise we should be using sys.path
    pip_env_future = asyncio.create_task(
        scan_pip(locations=locations or [Path(p) for p in sys.path])
    )
    conda_env = await conda_env_future
    pip_env = await pip_env_future
    filterd_conda = {}
    # the pip list is the "truth" of what is imported for python deps
    for name, package in conda_env.items():
        # if a package exists in the pip list but is not a conda place holder
        # then the conda package wont be imported and should be discarded
        if pip_env.get(name):
            if isinstance(pip_env[name], CondaPlaceHolder):
                filterd_conda[name] = package
            elif pip_env[name]["version"] == package["version"]:
                pip_env.pop(name, None)
                filterd_conda[name] = package
        else:
            # a non python package and safe to include
            filterd_conda[name] = package
    # remove conda placeholders
    pip_env = {
        pkg_name: pkg
        for pkg_name, pkg in pip_env.items()
        if not isinstance(pkg, CondaPlaceHolder)
    }
    return sorted(
        list(pip_env.values()) + list(filterd_conda.values()),
        key=lambda pkg: pkg["name"],
    )



import asyncio

import click
import rich

from ..utils import handle_credentials
from .utils import CONTEXT_SETTINGS


@click.command(context_settings=CONTEXT_SETTINGS)
@click.option("-s", "--server", help="Coiled server to use")
@click.option("-t", "--token", help="Coiled user token")
@click.option("-a", "--account", help="Coiled account")
@click.option(
    "--retry/--no-retry",
    default=True,
    help="Whether or not to automatically ask for a new token if an invalid token is entered",
)
def login(server, token, account, retry):
    """Configure your Coiled account credentials"""
    try:
        asyncio.get_event_loop().run_until_complete(
            handle_credentials(
                server=server, token=token, account=account, save=True, retry=retry
            )
        )
    except ImportError as e:
        rich.print(f"[red]{e}")



import os
import warnings
from typing import Optional, Tuple

import click

from ..core import create_software_environment
from ..utils import ExperimentalFeatureWarning
from .utils import (
    CONTEXT_SETTINGS,
    conda_command,
    conda_package_versions,
    parse_conda_command,
)


@click.command(
    context_settings=CONTEXT_SETTINGS,
    help="[EXPERIMENTAL] Create a Coiled software environment from a local conda environment",
)
@click.option(
    "-n",
    "--name",
    help="Name of conda environment to upload. Defaults to the current conda environment.",
)
def upload(name: Optional[str] = None):
    """Create a Coiled software environment from a local conda environment

    Parameters
    ----------
    name
        Name of conda environment to upload. Defaults to the current conda environment.

    Examples
    --------
    >>> import coiled
    >>> coiled.upload(name="my-local-env")

    """
    warnings.warn(
        "Uploading local software environments to Coiled is an experimental feature which is "
        "subject to breaking changes without warning",
        ExperimentalFeatureWarning,
    )
    # If not specified, use currently activate conda environment
    if name is None and "CONDA_DEFAULT_ENV" in os.environ:
        name = os.environ.get("CONDA_DEFAULT_ENV")
    if name is None:
        raise ValueError(
            "The local conda environment to upload cannot be determined. "
            "Please use the --name option to specify a local conda environment "
            "or activate the environment you want to upload before "
            "running 'coiled upload'."
        )
    # Get name and version of requested packages in the current conda environment
    requested_versions, channels = conda_export_requested(name=name)
    installed_versions = conda_package_versions(name=name)
    dependencies = []
    for package, version in requested_versions.items():
        if version is None:
            # No package version was requested, so use currently installed version
            version = f"={installed_versions[package]}"
        dependencies.append(f"{package}{version}")
    spec = {"channels": channels, "dependencies": dependencies}
    print(f"Creating Coiled software environment from local {name} conda environment")
    create_software_environment(name=name, conda=spec)


def conda_export_requested(name: str) -> Tuple[dict, list]:
    """Returns information from explicit specs in conda environment history

    Parameters
    ----------
    name
        Name of conda environment

    Returns
    -------
    versions
        Mapping that contains the name and version of each explicitly requested
        package in the environment
    channels
        List of conda channels from which packages were installed
    """
    cmd = [conda_command(), "env", "export", "-n", name, "--from-history", "--json"]
    output = parse_conda_command(cmd)
    deps = output.get("dependencies", [])
    versions: dict = {}
    for i in deps:
        if "=" not in i:
            # Version not specified
            versions[i] = None
        elif "version=" in i:
            # Version range specified. Example: django[version='>=3.0,<3.1']
            s = "version="
            start = i.find(s) + len(s) + 1
            stop = i.find("]") - 1
            version = i[start:stop]
            name, _ = i.split("[")
            versions[name] = version
        else:
            # Specific version specified. Example: dask==2.19.0
            info = i.split("=")
            # Remove empty strings originating from mutliple "="
            info = [i for i in info if i]
            name, version = info
            versions[name] = f"={version}"

    return versions, output["channels"]



import traceback

import click
from rich import print
from rich.prompt import Prompt

import coiled

from ..utils import CONTEXT_SETTINGS
from .util import setup_failure


@click.command(context_settings=CONTEXT_SETTINGS)
def setup_wizard() -> bool:
    return do_setup_wizard()


def do_setup_wizard() -> bool:
    coiled.add_interaction(
        action="cli-setup-wizard:start",
        success=True,
    )

    print(
        "Coiled creates and manages Dask clusters in your own cloud provider account.\n\n"
        "  1. AWS\n"
        "  2. Google Cloud\n"
        "  3. Azure\n"
        "  4. I don't have an AWS, GCP, or Azure account, help me choose!\n"
        "  [red]x[/red]. Exit the setup wizard\n"
    )

    try:
        choice = Prompt.ask(
            "Please choose your cloud provider so we can guide you in setting up Coiled",
            choices=["1", "2", "3", "4", "x"],
            show_choices=False,
        )
    except KeyboardInterrupt:
        coiled.add_interaction(
            action="cli-setup-wizard:KeyboardInterrupt", success=False
        )
        return False

    coiled.add_interaction(
        action="cli-setup-wizard:prompt", success=True, choice=choice
    )

    if choice == "1":  # AWS
        print("\nRunning [green]coiled setup aws[/green]\n")
        from .aws import do_setup

        try:
            return do_setup(slug="coiled")
        except Exception:
            setup_failure(f"Exception raised {traceback.format_exc()}", backend="aws")
            raise
    elif choice == "2":  # GCP
        print("\nRunning [green]coiled setup gcp[/green]\n")
        from .gcp import do_setup

        try:
            return do_setup()
        except Exception:
            setup_failure(f"Exception raised {traceback.format_exc()}", backend="gcp")
            raise
    elif choice == "3":  # Azure
        print(
            "\nWe don't currently offer managed infrastructure for Dask on Azure. If you're running Dask on Azure, "
            "Coiled can provide you with a platform for advanced analytics of our Dask usage. "
            "Please see our documentation about deploying Dask on Azure:\n"
            "[link]https://docs.coiled.io/user_guide/azure_reference.html[/link]"
        )
    elif choice == "4":  # Other
        print(
            "\nCoiled currently supports AWS and Google Cloud. It's easy to make an account with either and get "
            "started using Coiled. Please see our documentation about choosing a cloud provider:\n"
            "[link]https://docs.coiled.io/user_guide/backends.html#need-a-cloud-provider-account[/link]"
        )

    return False



import shutil

import pytest
from click.testing import CliRunner

import coiled
from coiled.cli.upload import upload

from ...utils import ExperimentalFeatureWarning
from ..utils import conda_command, parse_conda_command

pytestmark = pytest.mark.skipif(
    shutil.which("conda") is None,
    reason="skipping k8s tests because TEST_AGAINST_K8S is False",
)


@pytest.mark.test_group("test-upload")
def test_upload(sample_user):
    name = "coiled-test-foo"
    fqn = f"{sample_user.account.name}/{name}"
    parse_conda_command(
        [conda_command(), "create", "-y", "-q", "--name", name, "--json", "toolz"]
    )

    assert fqn not in coiled.list_software_environments()

    runner = CliRunner()
    with pytest.warns(ExperimentalFeatureWarning):
        result = runner.invoke(upload, args=f"--name {name}")
    assert result.exit_code == 0

    assert fqn in coiled.list_software_environments()



import asyncio
import datetime
from typing import Dict

from dask.utils import format_bytes, format_time
from rich.align import Align
from rich.console import RenderableType
from rich.markdown import Markdown
from rich.panel import Panel
from rich.rule import Rule
from rich.table import Table
from textual import events
from textual.app import App
from textual.reactive import Reactive
from textual.widget import Widget
from textual.widgets import ScrollView

from coiled._beta import CloudBeta
from coiled.core import Async

STATUS_ICON = {
    "running": "🟢",
    "stopped": "🔴",
}


class ClusterTable(Widget):
    selected_index: Reactive[int] = Reactive(0)
    has_focus: Reactive[bool] = Reactive(False)
    data: Reactive[Dict[str, Dict]] = Reactive({})

    def render(self):
        table = Table(
            expand=True,
            box=None,
            show_edge=True,
        )
        table.add_column("", no_wrap=True)
        table.add_column("id", no_wrap=True)
        table.add_column("created", no_wrap=True)
        table.add_column("workers", no_wrap=True)
        table.add_column("memory", no_wrap=True)
        table.add_column("n_tasks", no_wrap=True)
        table.add_column("time")
        table.add_column("$")
        for i, (name, dask_scheduler) in enumerate(self.data.items()):
            running = dask_scheduler["status"] == "running"
            table.add_row(
                STATUS_ICON[dask_scheduler["status"]],
                name.split("-")[1],
                dask_scheduler["created"].split("T")[0],
                str(dask_scheduler["workers"]) if running else "",
                format_bytes(dask_scheduler["memory"])
                + " / "
                + format_bytes(dask_scheduler["memory_limit"])
                if running
                else "",
                str(dask_scheduler["n_tasks"]),
                format_time(dask_scheduler["thread_seconds"]),
                "$%.2f" % (0.06 * dask_scheduler["thread_seconds"] / 3600),
                style="reverse" if self.selected_index == i else None,
            )

        return table


class Computations(Widget):
    data: Reactive[list] = Reactive([])

    def render(self):
        table = Table(
            expand=True,
            box=None,
            show_edge=True,
        )
        table.add_column("Code")
        table.add_column("Duration")
        for computation in self.data:
            start = datetime.datetime.fromisoformat(computation["start"])
            stop = datetime.datetime.fromisoformat(computation["stop"])
            table.add_row(
                Markdown("```python\n" + computation["code"] + "\n```"),
                format_time((stop - start).total_seconds()),
            )

        return table


class CoiledHeader(Widget):
    def render(self) -> RenderableType:
        return Panel(
            Align.center("Coiled", vertical="middle"),
            highlight=True,
            border_style="blue",
        )


class MyRule(Widget):
    def __init__(self, text):
        self.text = text
        Widget.__init__(self, name=text)

    def render(self):
        return Rule(self.text)


class CoiledUI(App):
    async def on_load(self) -> None:
        """Sent before going in to application mode."""
        await self.bind("q", "quit", "Quit")
        # TODO: Figure out why pyright is failing in this next line
        self.c: CloudBeta[Async] = await CloudBeta(asynchronous=True)  # type: ignore
        self.cluster_list = ClusterTable(name="Clusters")
        self.cluster_view = ScrollView(self.cluster_list, fluid=True)
        self.computations_list = Computations(name="Computations")
        self.computations_view = ScrollView(self.computations_list, fluid=True)

        await self._update_clusters()
        asyncio.create_task(self.update_clusters())
        asyncio.create_task(self.update_cluster())

    async def update_clusters(self) -> None:
        while True:
            await self._update_clusters()
            await asyncio.sleep(1)

    async def update_cluster(self) -> None:
        while True:
            await self._update_cluster()
            await asyncio.sleep(1)

    async def _update_clusters(self) -> None:
        clusters, _ = await self.c._list_dask_scheduler_page(0)
        clusters = {c["scheduler_id"]: c for c in clusters}
        self.cluster_list.data = clusters
        await self.cluster_view.update(self.cluster_list, home=False)

    async def _update_cluster(self):
        i = self.cluster_list.selected_index
        cluster = list(self.cluster_list.data.values())[i]
        # TODO: Figure out why pyright is failing in this next line
        self.computations_list.data = await self.c._list_computations(cluster["id"])  # type: ignore

    async def close_cluster(self):
        i = self.cluster_list.selected_index
        cluster = list(self.cluster_list.data.values())[i]
        await self.c._send_state(cluster["id"], desired_status="stopped")

    async def on_mount(self) -> None:
        await self.view.dock(CoiledHeader(), edge="top", name="header", size=3)
        await self.view.dock(
            self.cluster_view,
            size=10,
            edge="top",
            name="cluster_list",
        )
        await self.view.dock(
            MyRule("Computations"), edge="top", name="computations", size=3
        )

        await self.view.dock(
            self.computations_view,
            edge="top",
            name="computations",
        )

    async def on_key(self, event: events.Key) -> None:
        i = self.cluster_list.selected_index
        if event.key == "up":
            self.cluster_list.selected_index = (i - 1) % len(self.cluster_list.data)
        elif event.key == "down":
            self.cluster_list.selected_index = (i + 1) % len(self.cluster_list.data)
        elif event.key == "enter":
            await self._update_cluster()
        elif event.key == "delete":
            await self.close_cluster()


def show():
    app = CoiledUI()
    app.run(log="coiled.log")


if __name__ == "__main__":
    show()



import numpy as np
import optuna
import sklearn.datasets
import sklearn.metrics
import xgboost as xgb
from sklearn.model_selection import train_test_split


def objective(trial):
    # Load our dataset
    X, y = sklearn.datasets.load_breast_cancer(return_X_y=True)
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.25)
    dtrain = xgb.DMatrix(X_train, label=y_train)
    dtest = xgb.DMatrix(X_test, label=y_test)

    # Get set of hyperparameters
    param = {
        "silent": 1,
        "objective": "binary:logistic",
        "booster": trial.suggest_categorical("booster", ["gbtree", "dart"]),
        "lambda": trial.suggest_float("lambda", 1e-8, 1.0, log=True),
        "alpha": trial.suggest_float("alpha", 1e-8, 1.0, log=True),
        "max_depth": trial.suggest_int("max_depth", 1, 9),
        "eta": trial.suggest_float("eta", 1e-8, 1.0, log=True),
        "gamma": trial.suggest_float("gamma", 1e-8, 1.0, log=True),
        "grow_policy": trial.suggest_categorical(
            "grow_policy", ["depthwise", "lossguide"]
        ),
    }

    # Train XGBoost model
    bst = xgb.train(param, dtrain)
    preds = bst.predict(dtest)

    # Compute and return model accuracy
    pred_labels = np.rint(preds)
    accuracy = sklearn.metrics.accuracy_score(y_test, pred_labels)
    return accuracy


import dask_optuna
import joblib
from dask.distributed import Client

import coiled

# Create a Dask cluster with Coiled
cluster = coiled.Cluster(n_workers=5, software="optuna-example", name="optuna-example")
# Connect Dask to our cluster
client = Client(cluster)
print(f"Dask dashboard is available at {client.dashboard_link}")

# Create Dask-compatible Optuna storage class
storage = dask_optuna.DaskStorage()

# Run 500 optimizations trial on our cluster
study = optuna.create_study(direction="maximize", storage=storage)
with joblib.parallel_backend("dask"):
    study.optimize(objective, n_trials=500, n_jobs=-1)

# once your computation is done, close the cluster
cluster.close()
# Close the client
client.close()



import asyncio
import logging
import platform
import typing
from logging import getLogger
from pathlib import Path
from tempfile import TemporaryDirectory
from threading import Lock
from typing import Dict, List, Optional, Set

from packaging.utils import parse_wheel_filename

from coiled._beta.core import CloudBeta
from coiled.scan import PackageInfo, scan_prefix
from coiled.types import ResolvedPackageInfo
from coiled.utils import validate_wheel

logger = getLogger("coiled.package_sync")
subdir_datas = {}
PYTHON_VERSION = platform.python_version_tuple()


async def default_python() -> PackageInfo:
    python_version = platform.python_version()
    python_pkg: PackageInfo = {
        "name": "python",
        "path": None,
        "source": "conda",
        "channel_url": "https://conda.anaconda.org/conda-forge",
        "channel": "conda-forge",
        "subdir": "linux-64",
        "conda_name": "python",
        "version": python_version,
        "wheel_target": None,
    }
    return python_pkg


class PackageBuildError(Exception):
    pass


WHEEL_BUILD_LOCKS: Dict[str, Lock] = {}


async def create_wheel(pkg_name: str, version: str, src: str) -> ResolvedPackageInfo:
    lock = WHEEL_BUILD_LOCKS.setdefault(pkg_name, Lock())
    with lock:
        tmpdir = TemporaryDirectory()
        outdir = Path(tmpdir.name) / Path(pkg_name)
        logger.info(f"Attempting to create a wheel for {pkg_name} @ {src}")
        p = await asyncio.create_subprocess_shell(
            cmd=f"pip wheel --wheel-dir {outdir} --no-deps --use-pep517 --no-cache-dir {src}",
        )
        await p.wait()
        if p.returncode:
            return {
                "name": pkg_name,
                "source": "pip",
                "channel": None,
                "conda_name": None,
                "client_version": version,
                "specifier": "",
                "include": False,
                "error": (
                    "Failed to build a wheel for the"
                    " package, will not be included in environment, check stdout for details"
                ),
                "note": None,
                "sdist": None,
                "md5": None,
            }
        wheel_fn = next(file for file in outdir.iterdir() if file.suffix == ".whl")
        _, build_version, _, _ = parse_wheel_filename(str(wheel_fn.name))
        has_python, md5 = await validate_wheel(wheel_fn)
    return {
        "name": pkg_name,
        "source": "pip",
        "channel": None,
        "conda_name": None,
        "client_version": str(build_version),
        "specifier": "",
        "include": True,
        "error": None if has_python else "Built wheel contains no python files!",
        "note": f"Wheel built from {src}",
        "sdist": wheel_fn.open("rb"),
        "md5": md5,
    }


async def create_wheel_from_egg(
    pkg_name: str, version: str, src: str
) -> ResolvedPackageInfo:
    tmpdir = TemporaryDirectory()
    outdir = Path(tmpdir.name) / Path(pkg_name)
    outdir.mkdir(parents=True)
    logger.info(f"Attempting to create a wheel for {pkg_name} in directory {src}")
    p = await asyncio.create_subprocess_shell(
        cmd=f"wheel convert --dest-dir {outdir} {src}"
    )
    await p.wait()
    if p.returncode:
        return {
            "name": pkg_name,
            "source": "pip",
            "channel": None,
            "conda_name": None,
            "client_version": version,
            "specifier": "",
            "include": False,
            "error": (
                "Failed to build a wheel for the"
                " package, will not be included in environment, check stdout for details"
            ),
            "note": None,
            "sdist": None,
            "md5": None,
        }
    wheel_fn = next(file for file in outdir.iterdir() if file.suffix == ".whl")
    has_python, md5 = await validate_wheel(Path(wheel_fn))
    return {
        "name": pkg_name,
        "source": "pip",
        "channel": None,
        "conda_name": None,
        "client_version": version,
        "specifier": "",
        "include": True,
        "error": None if has_python else "Built wheel has no python files!",
        "note": "Wheel built from local egg",
        "sdist": wheel_fn.open("rb"),
        "md5": md5,
    }


async def approximate_packages(
    cloud: CloudBeta,
    packages: List[PackageInfo],
    priorities: Dict[str, int],
    strict: bool = False,
) -> typing.List[ResolvedPackageInfo]:
    results = await cloud._approximate_packages(
        packages=[
            {
                "name": pkg["name"],
                "priority_override": 100 if strict else priorities.get(pkg["name"]),
                "python_major_version": PYTHON_VERSION[0],
                "python_minor_version": PYTHON_VERSION[1],
                "python_patch_version": PYTHON_VERSION[2],
                "source": pkg["source"],
                "channel_url": pkg["channel_url"],
                "channel": pkg["channel"],
                "subdir": pkg["subdir"],
                "conda_name": pkg["conda_name"],
                "version": pkg["version"],
                "wheel_target": pkg["wheel_target"],
            }
            for pkg in packages
        ]
    )
    result_map = {r["name"]: r for r in results}
    finalized_packages: typing.List[ResolvedPackageInfo] = []
    for pkg in packages:
        if pkg["wheel_target"] and result_map[pkg["name"]]["include"]:
            if pkg["wheel_target"].endswith(".egg"):
                finalized_packages.append(
                    await create_wheel_from_egg(
                        pkg_name=pkg["name"],
                        version=pkg["version"],
                        src=pkg["wheel_target"],
                    )
                )
            else:
                finalized_packages.append(
                    await create_wheel(
                        pkg_name=pkg["name"],
                        version=pkg["version"],
                        src=pkg["wheel_target"],
                    )
                )
        else:
            finalized_packages.append(
                {
                    "name": pkg["name"],
                    "source": pkg["source"],
                    "channel": pkg["channel"],
                    "conda_name": pkg["conda_name"],
                    "client_version": pkg["version"],
                    "specifier": result_map[pkg["name"]]["specifier"] or "",
                    "include": result_map[pkg["name"]]["include"],
                    "note": result_map[pkg["name"]]["note"],
                    "error": result_map[pkg["name"]]["error"],
                    "sdist": None,
                    "md5": None,
                }
            )
    return finalized_packages


async def create_environment_approximation(
    cloud: CloudBeta,
    priorities: Dict[str, int],
    only: Optional[Set[str]] = None,
    strict: bool = False,
) -> typing.List[ResolvedPackageInfo]:
    packages = await scan_prefix()
    python = next((p for p in packages if p["name"] == "python"), None)
    if not python:
        packages.append(await default_python())
    if only:
        packages = filter(lambda pkg: pkg["name"] in only, packages)
    # TODO: private conda channels
    result = await approximate_packages(
        cloud=cloud,
        packages=[pkg for pkg in packages],
        priorities=priorities,
        strict=strict,
    )
    return result


if __name__ == "__main__":
    from logging import basicConfig

    basicConfig(level=logging.INFO)

    from rich.console import Console
    from rich.table import Table

    async def run():
        async with CloudBeta(asynchronous=True) as cloud:
            return await create_environment_approximation(
                cloud=cloud,
                priorities={"dask": 100, "twisted": -2, "graphviz": -1, "icu": -1},
            )

    result = asyncio.run(run())

    table = Table(title="Packages")
    keys = ("name", "source", "include", "client_version", "specifier", "error", "note")

    for key in keys:
        table.add_column(key)

    for pkg in result:
        row_values = [str(pkg.get(key, "")) for key in keys]
        table.add_row(*row_values)
    console = Console()
    console.print(table)



import click

from ..utils import CONTEXT_SETTINGS
from .amp import aws_amp_setup
from .aws import aws_setup
from .entry import setup_wizard
from .gcp import gcp_setup


@click.group(context_settings=CONTEXT_SETTINGS)
def setup():
    """Setup Coiled with cloud provider"""
    pass


setup.add_command(setup_wizard, "wizard")
setup.add_command(aws_setup, "aws")
setup.add_command(gcp_setup, "gcp")

setup.add_command(aws_amp_setup, "amp")



import dask
import dask.dataframe as dd
import folium
import streamlit as st
from dask.distributed import Client
from folium.plugins import HeatMap
from streamlit_folium import folium_static

import coiled

# Text in Streamlit
st.header("Coiled and Streamlit")
st.subheader("Analyzing Large Datasets with Coiled and Streamlit")
st.write(
    """
    The computations for this Streamlit app are powered by Coiled, which
    provides on-demand, hosted Dask clusters in the cloud. Change the options
    below to view different visualizations of transportation pickups/dropoffs,
    then let Coiled handle all of the infrastructure and compute.
    """
)

# Interactive widgets in Streamlit
taxi_mode = st.selectbox("Taxi pickup or dropoff?", ("Pickups", "Dropoffs"))
num_passengers = st.slider("Number of passengers", 0, 9, (0, 9))

# Start and connect to Coiled cluster
cluster_state = st.empty()


@st.cache(allow_output_mutation=True)
def get_client():
    cluster_state.write("Starting or connecting to Coiled cluster...")
    cluster = coiled.Cluster(
        n_workers=5, name="coiled-streamlit", software="streamlit-example"
    )
    client = Client(cluster)
    return client


client = get_client()
if client.status == "closed":
    # In a long-running Streamlit app,
    # the cluster could have shut down from idleness.
    # If so, clear the Streamlit cache to restart it.
    st.caching.clear_cache()
    client = get_client()
cluster_state.write(f"Coiled cluster is up! ({client.dashboard_link})")


# Load data (runs on Coiled)
@st.cache(hash_funcs={dd.DataFrame: dask.base.tokenize})
def load_data():
    return (
        dd.read_parquet(
            "s3://coiled-datasets/nyc-tlc/2019",
            usecols=[
                "passenger_count",
                "pickup_longitude",
                "pickup_latitude",
                "dropoff_longitude",
                "dropoff_latitude",
            ],
            storage_options={"anon": True},
            blocksize="16 MiB",
        )
        .dropna()
        .persist()
    )


df = load_data()

# Filter data based on inputs (runs on Coiled)
with st.spinner("Calculating map data..."):
    map_data = df[
        (df["passenger_count"] >= num_passengers[0])
        & (df["passenger_count"] <= num_passengers[1])
    ]

    if taxi_mode == "Pickups":
        map_data = map_data.iloc[:, [2, 1]]
    elif taxi_mode == "Dropoffs":
        map_data = map_data.iloc[:, [4, 3]]

    map_data.columns = ["lat", "lon"]
    map_data = map_data.loc[~(map_data == 0).all(axis=1)]
    map_data = map_data.head(500)

# Display map in Streamlit
st.subheader("Map of selected rides")
m = folium.Map([40.76, -73.95], tiles="cartodbpositron", zoom_start=12)
HeatMap(map_data).add_to(folium.FeatureGroup(name="Heat Map").add_to(m))
folium_static(m)
